#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <odbx.h>
#include "odbx-regression.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif



struct odbxtest
{
	char* name;
	int value;
	int result;
	int try;
};


static struct odbxtest cap[] = {
	{ "ODBX_CAP_BASIC", ODBX_CAP_BASIC, 0, 1 },
};

#define CAPMAX 1


static struct odbxtest opt[] = {
	{ "ODBX_OPT_API_VERSION", ODBX_OPT_API_VERSION, 0, 0 },
	{ "ODBX_OPT_THREAD_SAFE", ODBX_OPT_THREAD_SAFE, 0, 0 },
	{ "ODBX_OPT_TLS", ODBX_OPT_TLS, 0, 0 },
	{ "ODBX_OPT_MULTI_STATEMENTS", ODBX_OPT_MULTI_STATEMENTS, 0, 1 },
	{ "ODBX_OPT_PAGED_RESULTS", ODBX_OPT_PAGED_RESULTS, 0, 1 },
	{ "ODBX_OPT_COMPRESS", ODBX_OPT_COMPRESS, 0, 1 }
};

#define OPTMAX 6



int exec( odbx_t* handle[], struct odbxstmt qptr[], int verbose );



void help( char* name )
{
	fprintf( stdout, "Usage: %s <options>\n", name );
	fprintf( stdout, "   -b <backend>    Database backend (mysql, pgsql, sqlite, sqlite3, firebird, freetds, sybase)\n" );
	fprintf( stdout, "   -h <server>     Server name, IP address or directory\n" );
	fprintf( stdout, "   -p <port>       Server port\n" );
	fprintf( stdout, "   -d <database>   Database or file name\n" );
	fprintf( stdout, "   -u <user>       User name for authentication\n" );
	fprintf( stdout, "   -w <password>   Password for authentication\n" );
	fprintf( stdout, "   -e              Force encrypted connection\n" );
	fprintf( stdout, "   -r              Number of runs\n" );
	fprintf( stdout, "   -v              Verbose mode\n" );
}



int main( int argc, char* argv[] )
{
	int i, j, k, err = 0;
	struct odbxstmt** stmts = NULL;
	struct odbxstmt* queries = NULL;
	odbx_t* handle[2] = { NULL, NULL };
	char *backend, *host, *port, *db, *user, *pass;
	int param, verbose = 0, encrypt = 0, runs = 1;


	backend = host = port = db = user = pass = NULL;

	while( ( param = getopt( argc, argv, "b:h:p:d:u:w:e:r:v" ) ) != -1 )
	{
		switch( param )
		{
			case 'b':
				backend = optarg;
				break;
			case 'h':
				host = optarg;
				break;
			case 'p':
				port = optarg;
				break;
			case 'd':
				db = optarg;
				break;
			case 'u':
				user = optarg;
				break;
			case 'w':
				pass = optarg;
				break;
			case 'e':
				if( optarg != NULL ) { encrypt = (int) strtol( optarg, NULL, 10 ); }
				break;
			case 'v':
				verbose = 1;
				break;
			case 'r':
				if( optarg != NULL ) { runs = (int) strtol( optarg, NULL, 10 ); }
				break;
			default:
				fprintf( stderr, "Unknown option '%c' with arg '%s'\n", (char) param, optarg );
				return 1;
		}
	}

	if( backend == NULL ) { help( argv[0] ); return 1; }

	if( !strncmp( backend, "firebird", 8 ) ) { stmts = firebird_stmt; }
	if( !strncmp( backend, "mssql", 5 ) ) { stmts = mssql_stmt; }
	if( !strncmp( backend, "mysql", 5 ) ) { stmts = mysql_stmt; }
	if( !strncmp( backend, "oracle", 6 ) ) { stmts = oracle_stmt; }
	if( !strncmp( backend, "pgsql", 5 ) ) { stmts = pgsql_stmt; }
	if( !strncmp( backend, "sqlite", 6 ) ) { stmts = sqlite_stmt; }
	if( !strncmp( backend, "sqlite3", 7 ) ) { stmts = sqlite3_stmt; }
	if( !strncmp( backend, "sybase", 6 ) ) { stmts = sybase_stmt; }

	if( stmts == NULL )
	{
		fprintf( stderr, "Invalid backend: %s\n", backend );
		return 1;
	}
	queries = stmts[0];


	if( verbose )
	{
		fprintf( stdout, "\nParameter:\n" );
		if( backend ) { fprintf( stdout, "  backend=%s\n", backend ); }
		if( host ) { fprintf( stdout, "  host=%s\n", host ); }
		if( port ) { fprintf( stdout, "  port=%s\n", port ); }
		if( db ) { fprintf( stdout, "  database=%s\n", db ); }
		if( user ) { fprintf( stdout, "  user=%s\n", user ); }
		if( pass ) { fprintf( stdout, "  password=%s\n", pass ); }
		if( encrypt ) { fprintf( stdout, "  encrypt=%d\n", encrypt ); }
		if( runs ) { fprintf( stdout, "  runs=%d\n", runs ); }
		if( verbose ) { fprintf( stdout, "  verbose=%d\n",verbose ); }
	}


	for( j = 0; j < runs; j++ )
	{
		if( verbose ) { fprintf( stdout, "\n%d. Run:\n", j+1 ); }

		for( k = 0; k < 2; k++ )
		{
			if( verbose ) { fprintf( stdout, "  odbx_init()\n" ); }
			if( ( err = odbx_init( &(handle[k]), backend, host, port ) ) < 0 )
			{
				fprintf( stderr, "Error in odbx_init(): %s\n", odbx_error( handle[k], err ) );
				goto FAIL;
			}

			if( verbose ) { fprintf( stdout, "  odbx_capabilities()\n" ); }
			for( i = 0; i < CAPMAX; i++ )
			{
				if( ( err = odbx_capabilities( handle[k], (unsigned int) cap[i].value ) ) < 0 )
				{
					fprintf( stderr, "Error in odbx_capabilities(): %s\n", odbx_error( handle[k], err ) );
					goto ERROR;
				}

				cap[i].result = err;
				fprintf( stdout, "    %s: %d\n", cap[i].name, cap[i].result );
			}

			if( verbose ) { fprintf( stdout, "  odbx_{get/set}_option()\n" ); }
			for( i = 0; i < OPTMAX; i++ )
			{
				if( ( err = odbx_get_option( handle[k], opt[i].value, (void*) &(opt[i].result) ) ) < 0 )
				{
					fprintf( stderr, "Error in odbx_get_option(): %s\n", odbx_error( handle[k], err ) );
					goto ERROR;
				}

				fprintf( stdout, "    %s: %d %s\n", opt[i].name, opt[i].result, opt[i].result && opt[i].try ? "(using)" : "" );

				if( opt[i].result && opt[i].try && ( err = odbx_set_option( handle[k], opt[i].value, (void*) &(opt[i].try) ) ) < 0 )
				{
					fprintf( stderr, "Error in odbx_set_option(): %s\n", odbx_error( handle[k], err ) );
					goto ERROR;
				}

				if( opt[i].value == ODBX_OPT_MULTI_STATEMENTS && opt[i].try && opt[i].result ) { queries = stmts[1]; }
			}

			if( encrypt && ( err = odbx_set_option( handle[k], ODBX_OPT_TLS, (void*) &encrypt ) ) < 0 )
			{
				fprintf( stderr, "Error in odbx_set_option(): %s\n", odbx_error( handle[k], err ) );
				goto ERROR;
			}

			if( verbose ) { fprintf( stdout, "  odbx_bind()\n" ); }
			if( ( err = odbx_bind( handle[k], db, user, pass, ODBX_BIND_SIMPLE ) ) < 0 )
			{
				fprintf( stderr, "Error in odbx_bind(): %s\n", odbx_error( handle[k], err ) );
				goto ERROR;
			}
		}

		for( i = 0; i < runs; i++ )
		{
			if( ( err = exec( handle, queries, verbose ) ) < 0 )
			{
				fprintf( stdout, "Error in exec(): Fatal error in run %d\n", i );
				break;
			}
		}

		for( k = 0; k < 2; k++ )
		{
			if( verbose ) { fprintf( stdout, "  odbx_unbind()\n" ); }
			if( odbx_unbind( handle[k] ) < 0 )
			{
					fprintf( stderr, "Error in odbx_unbind(): %s\n", odbx_error( handle[k], err ) );
			}
		}

ERROR:

		for( k = 0; k < 2; k++ )
		{
			if( verbose && handle[k] ) { fprintf( stdout, "  odbx_finish()\n" ); }
			if( handle[k] && odbx_finish( handle[k] ) < 0 )
			{
					fprintf( stderr, "Error in odbx_finish(): %s\n", odbx_error( handle[k], err ) );
			}
		}
	}

FAIL:

	return err;
}


int exec( odbx_t* handle[], struct odbxstmt* qptr, int verbose )
{
	char* tmp;
	int i, err = 0;
	struct timeval tv;
	odbx_result_t* result;
	unsigned long fields;


	while( qptr->str != NULL )
	{
		if( verbose ) { fprintf( stdout, "  odbx_query(%d): '%s'\n", qptr->num, qptr->str ); }
		if( ( err = odbx_query( handle[qptr->num], qptr->str, strlen( qptr->str ) ) ) < 0 )
		{
			fprintf( stderr, "Error in odbx_query(): %s\n", odbx_error( handle[qptr->num], err ) );
			if( odbx_error_type( handle[qptr->num], err ) < 0 ) { return err; }
			qptr++;
			continue;
		}

		tv.tv_sec = 3;
		tv.tv_usec = 0;

		/* use values >1 for paged output (if supported) or 0 for all rows */
		while( ( err = odbx_result( handle[qptr->num], &result, &tv, 5 ) ) != 0 )
		{
			if( verbose ) { fprintf( stdout, "  odbx_result()\n" ); }

			tv.tv_sec = 3;
			tv.tv_usec = 0;

			if( err < 0 )
			{
				fprintf( stderr, "Error in odbx_result(): %s\n", odbx_error( handle[qptr->num], err ) );
				if( odbx_error_type( handle[qptr->num], err ) < 0 ) { return err; }
				continue;
			}

			switch( err )
			{
				case 1:
					fprintf( stdout, "    odbx_result(): Timeout\n" );
					continue;
				case 2:
					fprintf( stdout, "    Affected rows: %lld\n", odbx_rows_affected( result ) );
					odbx_result_free( result );
					continue;
			}

			if( verbose ) { fprintf( stdout, "  odbx_column_count()\n" ); }
			fields = odbx_column_count( result );

			while( ( err = odbx_row_fetch( result ) ) != 0 )
			{
				if( verbose ) { fprintf( stdout, "  odbx_row_fetch()\n" ); }

				if( err < 0 )
				{
					fprintf( stderr, "Error in odbx_row_fetch(): %s\n", odbx_error( handle[qptr->num], err ) );
					if( odbx_error_type( handle[qptr->num], err ) < 0 ) { return err; }
					continue;
				}

				for( i = 0; i < fields; i++ )
				{
					fprintf( stdout, "    column %d (type %d): ", i, odbx_column_type( result, i ) );

					tmp = (char*) odbx_column_name( result, i );
					if( tmp != NULL ) { fprintf( stdout, "%s", tmp ); }

					tmp = (char*) odbx_field_value( result, i );
					if( tmp != NULL ) { fprintf( stdout, " = '%s'\n", tmp ); }
					else { fprintf( stdout, " = NULL\n" ); }
				}
			}

			if( verbose ) { fprintf( stdout, "  odbx_result_free()\n" ); }
			odbx_result_free( result );
		}

		qptr++;
	}

	return 0;
}
